@@ -44,6 +44,7 @@ model = Transfusion(
     modality_decoder = Decoder(),
     add_pos_emb = True,
     modality_num_dim = 2,
+    velocity_consistency_loss_weight = 0.1,
     transformer = dict(
         dim = 64,
         depth = 4,
@@ -52,6 +53,8 @@ model = Transfusion(
     )
 ).cuda()
 
+ema_model = model.create_ema()
+
 class MnistDataset(Dataset):
     def __init__(self):
         self.mnist = torchvision.datasets.MNIST(
@@ -83,7 +86,7 @@ optimizer = Adam(model.parameters(), lr = 8e-4)
 
 for step in range(1, 100_000 + 1):
 
-    loss = model(next(iter_dl))
+    loss = model(next(iter_dl), velocity_consistency_ema_model = ema_model)
     loss.backward()
 
     torch.nn.utils.clip_grad_norm_(model.parameters(), 0.5)
@@ -91,10 +94,12 @@ for step in range(1, 100_000 + 1):
     optimizer.step()
     optimizer.zero_grad()
 
+    ema_model.update()
+
     print(f'{step}: {loss.item():.3f}')
 
     if divisible_by(step, 500):
-        image = model.generate_modality_only(batch_size = 64)
+        image = ema_model.generate_modality_only(batch_size = 64)
 
         save_image(
             rearrange(image, '(gh gw) 1 h w -> 1 (gh h) (gw w)', gh = 8).detach().cpu(),
