@@ -28,12 +28,12 @@ def divisible_by(num, den):
 
 class Encoder(Module):
     def forward(self, x):
-        x = rearrange(x, '1 (h p1) (w p2) -> h w (p1 p2)', p1 = 2, p2 = 2)
+        x = rearrange(x, '... 1 (h p1) (w p2) -> ... h w (p1 p2)', p1 = 2, p2 = 2)
         return x * 2 - 1
 
 class Decoder(Module):
     def forward(self, x):
-        x = rearrange(x, 'h w (p1 p2) -> 1 (h p1) (w p2)', p1 = 2, p2 = 2, h = 14)
+        x = rearrange(x, '... h w (p1 p2) -> ... 1 (h p1) (w p2)', p1 = 2, p2 = 2, h = 14)
         return ((x + 1) * 0.5).clamp(min = 0., max = 1.)
 
 model = Transfusion(
@@ -52,10 +52,12 @@ model = Transfusion(
     )
 ).cuda()
 
+ema_model = model.create_ema()
+
 class MnistDataset(Dataset):
     def __init__(self):
         self.mnist = torchvision.datasets.MNIST(
-            './data',
+            './data/mnist',
             download = True
         )
 
@@ -77,8 +79,8 @@ def collate_fn(data):
     return data
 
 dataset = MnistDataset()
+dataloader = model.create_dataloader(dataset, batch_size = 16, shuffle = True)
 
-dataloader = DataLoader(dataset, batch_size = 16, collate_fn = collate_fn, shuffle = True)
 iter_dl = cycle(dataloader)
 
 optimizer = Adam(model.parameters(), lr = 3e-4)
@@ -96,12 +98,14 @@ for step in range(1, 10_000 + 1):
     optimizer.step()
     optimizer.zero_grad()
 
+    ema_model.update()
+
     print(f'{step}: {loss.item():.3f}')
 
     # eval
 
     if divisible_by(step, 500):
-        one_multimodal_sample = model.sample(max_length = 10, fixed_modality_shape = (14, 14))
+        one_multimodal_sample = ema_model.sample(max_length = 10)
 
         print_modality_sample(one_multimodal_sample)
 
