@@ -16,13 +16,15 @@ p - positions
 import os
 
 import math
-from functools import partial
+from functools import partial, wraps
 from typing import NamedTuple, Callable, Literal
 
 import torch
-from torch import nn, Tensor, tensor, is_tensor
 import torch.nn.functional as F
+from torch import nn, Tensor, tensor, is_tensor, stack
 from torch.nn import Module, ModuleList, Linear
+
+from torch.utils.data import Dataset, DataLoader
 from torch.nn.utils.rnn import pad_sequence
 from torch.utils._pytree import tree_map, tree_flatten, tree_unflatten
 
@@ -85,7 +87,6 @@ class LossBreakdown(NamedTuple):
     text: Float['']
     flow: list[Float['']]
     velocity: list[Float['']] | None
-    direction: list[Float['']] | None
 
 # helper functions
 
@@ -113,6 +114,15 @@ def cast_tuple(t, length = 1):
 def tree_map_tensor(sample, fn: Callable):
     return tree_map(lambda t: t if not is_tensor(t) else fn(t), sample)
 
+def add_temp_batch_dim(fn: Callable):
+    @wraps(fn)
+    def inner(t: Tensor, *args, **kwargs) -> Tensor:
+        t = rearrange(t, '... -> 1 ...')
+        out = fn(t, *args, **kwargs)
+        out = rearrange(out, '1 ... -> ...')
+        return out
+    return inner
+
 def eval_decorator(fn):
     def inner(self, *args, **kwargs):
         was_training = self.training
@@ -226,6 +236,15 @@ def gumbel_sample(t, temperature = 1., dim = -1, keepdim = True):
     noise = gumbel_noise(t) * int(temperature > 0)
     return (t / temperature + noise).argmax(dim = dim, keepdim = keepdim)
 
+# dataloader related
+
+def collate_fn(data):
+    return [*map(list, data)]
+
+@typecheck
+def create_dataloader(dataset: Dataset, **kwargs) -> DataLoader:
+    return DataLoader(dataset, collate_fn = collate_fn, **kwargs)
+
 # flex attention mask construction
 # https://pytorch.org/blog/flexattention/
 
@@ -262,11 +281,6 @@ def softcap_score_mod(softcap):
         return score
     return inner
 
-# losses
-
-def calc_direction_loss(pred, target):
-    return 0.5 * (1. - einsum(l2norm(pred), l2norm(target), '... d, ... d -> ...'))
-
 # converting a raw list of modality offsets and lengths to tensor
 
 @typecheck
@@ -291,7 +305,7 @@ def order_modality_positions_by_seq_offset(
     modalities: Int['b m 3']
 ) -> tuple[Int['b m 3'], Int['b m']]:
 
-    type, offsets, lengths = modalities.unbind(dim = -1)
+    modality_type, offsets, lengths = modalities.unbind(dim = -1)
 
     no_modality_mask = lengths <= 0 # there may be uneven number of modalities per batch sample
     offsets_to_sort = offsets.masked_fill(no_modality_mask, 1e10)
@@ -334,9 +348,9 @@ def embed_modality_tokens(
     output = torch.zeros((batch, seq_len, dim), device = device)
 
     for batch_ind, (one_modality, one_modality_token) in enumerate(zip(modalities, modality_tokens)):
-        for (type, offset, length), batch_modality_token in zip(one_modality, one_modality_token):
+        for (modality_type, offset, length), batch_modality_token in zip(one_modality, one_modality_token):
 
-            if modality_id != type or length <= 0:
+            if modality_id != modality_type or length <= 0:
                 continue
 
             modality_shape = batch_modality_token.shape
@@ -458,7 +472,7 @@ class MLPAxialPositions(Module):
         dimensions = modality_shape.tolist()
 
         grid = torch.meshgrid([torch.arange(dim_len, device = self.device) for dim_len in dimensions], indexing = 'ij')
-        axial_positions = torch.stack(grid, dim = -1)
+        axial_positions = stack(grid, dim = -1)
 
         pos_emb = self.mlp(axial_positions.float())
 
@@ -741,7 +755,7 @@ class Attention(Module):
         # maybe kv cache
 
         if return_kv_cache:
-            kv_cache = torch.stack((k, v))
+            kv_cache = stack((k, v))
 
         # rotary embeddings
 
@@ -987,7 +1001,7 @@ class Transformer(Module):
         if not return_kv_cache:
             return out
 
-        return out, torch.stack(new_cache)
+        return out, stack(new_cache)
 
 # classes
 
@@ -1011,9 +1025,8 @@ class Transfusion(Module):
         ignore_index = -1,
         flow_loss_weight = 1.,
         text_loss_weight = 1.,
-        add_flow_direction_loss = False, # figure 7 https://arxiv.org/abs/2410.10356
-        direction_loss_weight = 1.,
-        velocity_consistency_loss_weight = 1.,
+        velocity_consistency_loss_weight = 0.1,
+        modality_encoder_decoder_requires_batch_dim = True, # whether the modality encoder / decoder requires batch dimension, will auto assume it is needed
         odeint_kwargs: dict = dict(
             atol = 1e-5,
             rtol = 1e-5,
@@ -1090,6 +1103,10 @@ class Transfusion(Module):
         assert len(self.modality_encoder) == self.num_modalities
         assert len(self.modality_decoder) == self.num_modalities
 
+        # auto handle batch dimension for modality encoder / decoder
+
+        self.maybe_add_temp_batch_dim = add_temp_batch_dim if modality_encoder_decoder_requires_batch_dim else identity
+
         # default token lengths for respective modality
         # fallback if the language model does not come up with valid dimensions
 
@@ -1166,19 +1183,50 @@ class Transfusion(Module):
 
         self.velocity_consistency_loss_weight = velocity_consistency_loss_weight
 
-        # flow direction loss related
-
-        self.add_flow_direction_loss = add_flow_direction_loss
-        self.direction_loss_weight = direction_loss_weight
-
         # flow sampling related
 
         self.odeint_fn = partial(odeint, **odeint_kwargs)
 
+        # dummy loss
+
+        self.register_buffer('zero', tensor(0.), persistent = False)
+
     @property
     def device(self):
         return next(self.parameters()).device
 
+    def parameters_without_encoder_decoder(self):
+        return (
+            set(self.parameters()) -
+            set(self.modality_encoder.parameters()) -
+            set(self.modality_decoder.parameters())
+        )
+
+    def create_dataloader(
+        self,
+        *args,
+        **kwargs
+    ):
+        return create_dataloader(*args, **kwargs)
+
+    def create_ema(
+        self,
+        beta = 0.99,
+        *ema_kwargs
+    ) -> EMA:
+
+        ema = EMA(
+            self,
+            beta = beta,
+            forward_method_names = (
+                'sample',
+                'generate_text_only',
+                'generate_modality_only'
+            )
+        )
+
+        return ema
+
     @torch.no_grad()
     @eval_decorator
     @typecheck
@@ -1412,7 +1460,8 @@ class Transfusion(Module):
                 modality = rearrange(modality, '... d -> d ...')
 
             if exists(maybe_modality_decoder):
-                modality = maybe_modality_decoder(modality)
+                maybe_modality_decoder.eval()
+                modality = self.maybe_add_temp_batch_dim(maybe_modality_decoder)(modality)
 
             processed_modality_sample.append((modality_id, modality))
 
@@ -1515,9 +1564,14 @@ class Transfusion(Module):
         times: Float['b'] | None = None,
         modality_type: int | None = None,
         encode_modality: bool = True,
-        return_loss = True
+        velocity_consistency_ema_model: Transfusion | None = None,
+        velocity_consistency_delta_time = 1e-5,
+        return_loss = True,
+        return_loss_breakdown = False
     ) -> Float[''] | Float['b ...']:
 
+        requires_velocity_consistency = exists(velocity_consistency_ema_model)
+
         modalities = modalities.to(self.device)
 
         if self.num_modalities > 1:
@@ -1541,7 +1595,12 @@ class Transfusion(Module):
         if encode_modality and exists(maybe_modality_encode):
             with torch.no_grad():
                 maybe_modality_encode.eval()
-                modalities = maybe_modality_encode(modalities).detach()
+                modalities = self.maybe_add_temp_batch_dim(maybe_modality_encode)(modalities).detach()
+
+        # maybe channel first
+
+        if self.channel_first_latent:
+            modalities = rearrange(modalities, 'b d ... -> b ... d')
 
         shape = modalities.shape
 
@@ -1552,10 +1611,7 @@ class Transfusion(Module):
         if add_pos_emb:
             assert exists(modality_num_dim), f'modality_num_dim must be set for modality {modality_type} if further injecting axial positional embedding'
 
-            if self.channel_first_latent:
-                _, _, *axial_dims = shape
-            else:
-                _, *axial_dims, _ = shape
+            _, *axial_dims, _ = shape
 
             assert len(axial_dims) == modality_num_dim, f'received modalities of ndim {len(axial_dims)} but expected {modality_num_dim}'
 
@@ -1563,23 +1619,25 @@ class Transfusion(Module):
 
         batch, device = tokens.shape[0], tokens.device
 
-        # maybe channel first
-
-        if self.channel_first_latent:
-            tokens = rearrange(tokens, 'b d ... -> b ... d')
-
         # times
 
         if not exists(times):
             times = torch.rand((batch,), device = device)
 
-        padded_times = append_dims(times, tokens.ndim - 1)
+        if return_loss:
 
-        noise = torch.randn_like(tokens)
+            if requires_velocity_consistency:
+                times = times * (1. - velocity_consistency_delta_time) # make sure times are max of 1. - small delta, for velocity consistency
 
-        noised_tokens = padded_times * tokens + (1. - padded_times) * noise
+            padded_times = append_dims(times, tokens.ndim - 1)
 
-        flow = tokens - noise
+            noise = torch.randn_like(tokens)
+
+            noised_tokens = padded_times * tokens + (1. - padded_times) * noise
+
+            flow = tokens - noise
+        else:
+            noised_tokens = tokens
 
         # from latent to model tokens
 
@@ -1616,7 +1674,36 @@ class Transfusion(Module):
 
         flow = rearrange(flow, 'b ... d -> b (...) d')
 
-        return F.mse_loss(pred_flow, flow)
+        flow_loss = F.mse_loss(pred_flow, flow)
+
+        # maybe velocity consistency loss
+
+        velocity_loss = self.zero
+
+        if requires_velocity_consistency:
+
+            with torch.no_grad():
+                flow_with_delta_time = velocity_consistency_ema_model.forward_modality(
+                    modalities = modalities,
+                    modality_type = modality_type,
+                    times = times + velocity_consistency_delta_time,
+                    encode_modality = False, # modality already encoded
+                    return_loss = False
+                )
+
+            velocity_loss = F.mse_loss(flow, flow_with_delta_time)
+
+        # total loss
+
+        total_loss = (
+            flow_loss +
+            velocity_loss * self.velocity_consistency_loss_weight
+        )
+
+        if not return_loss_breakdown:
+            return total_loss
+
+        return total_loss, (flow_loss, velocity_loss)
 
     @torch.no_grad()
     @eval_decorator
@@ -1684,6 +1771,7 @@ class Transfusion(Module):
             sampled_modality = rearrange(sampled_modality, 'b ... d -> b d ...')
 
         if exists(maybe_modality_decoder):
+            maybe_modality_decoder.eval()
             sampled_modality = maybe_modality_decoder(sampled_modality)
 
         return sampled_modality
@@ -1747,7 +1835,13 @@ class Transfusion(Module):
 
         if is_modality_only:
             assert return_loss
-            return self.forward_modality(modalities, modality_type = modality_type)
+
+            forward_modality_kwargs = dict(
+                modality_type = modality_type,
+                velocity_consistency_ema_model = velocity_consistency_ema_model
+            )
+
+            return self.forward_modality(modalities, **forward_modality_kwargs)
 
         device = self.device
         tensor_ = partial(tensor, device = device)
@@ -1799,12 +1893,18 @@ class Transfusion(Module):
                     modality = (0, modality)
 
                 is_text = not isinstance(modality, tuple)
+                is_modality = not is_text
 
                 if is_text:
                     modality_tensor = modality
                 else:
                     modality_type, modality_tensor = modality
 
+                # auto move modality tensor to correct device
+
+                modality_tensor = modality_tensor.to(device)
+
+                if is_modality:
                     if not is_decoding:
                         modality_transform = self.modality_token_transform[modality_type]
                         maybe_modality_encode = self.modality_encoder[modality_type]
@@ -1815,7 +1915,7 @@ class Transfusion(Module):
 
                             with torch.no_grad():
                                 maybe_modality_encode.eval()
-                                modality_tensor = maybe_modality_encode(modality_tensor).detach()
+                                modality_tensor = self.maybe_add_temp_batch_dim(maybe_modality_encode)(modality_tensor).detach()
 
                         if self.channel_first_latent:
                             modality_tensor = rearrange(modality_tensor, 'd ... -> ... d')
@@ -1823,9 +1923,7 @@ class Transfusion(Module):
                     assert 0 <= modality_type < self.num_modalities, f'received a modality index that is out of range. only {self.num_modalities} modalities specified'
                     assert self.dim_latents[modality_type] == modality_tensor.shape[-1], f'mismatch for modality latent dimension - expected {self.dim_latents[modality_type]} but received {modality_tensor.shape[-1]} - modality shape is {tuple(modality_tensor.shape)}, perhaps you need to set `channel_first_latent = True`'
 
-                # auto move modality tensor to device of model
-
-                modality_tensor = modality_tensor.to(device)
+                # auto ward against scalars (lone start end tokens)
 
                 if modality_tensor.dtype in (torch.int, torch.long) and modality_tensor.ndim == 0:
                     modality_tensor = rearrange(modality_tensor, '-> 1')
@@ -2101,7 +2199,6 @@ class Transfusion(Module):
         # calculate flow losses
 
         flow_losses = []
-        direction_losses = []
 
         modality_loss_weights = []
 
@@ -2117,36 +2214,21 @@ class Transfusion(Module):
 
             flow_loss = flow_loss[is_one_modality].mean()
 
-            if self.add_flow_direction_loss:
-                direction_loss = calc_direction_loss(pred_flow, flow)
-                direction_loss = direction_loss[is_one_modality].mean()
-
             modality_loss_weight = is_one_modality.sum() / total_tokens
 
             modality_loss_weights.append(modality_loss_weight)
 
             flow_losses.append(flow_loss)
 
-            if self.add_flow_direction_loss:
-                direction_losses.append(direction_loss)
-
-        modality_loss_weights = torch.stack(modality_loss_weights)
+        modality_loss_weights = stack(modality_loss_weights)
 
         # only the token positions that are not modalities have autoregressive loss
 
         total_loss = (
             text_loss * text_loss_weight * self.text_loss_weight +
-            (torch.stack(flow_losses) * modality_loss_weights).sum() * self.flow_loss_weight
+            (stack(flow_losses) * modality_loss_weights).sum() * self.flow_loss_weight
         )
 
-        # maybe add direction loss
-
-        if self.add_flow_direction_loss:
-            total_loss = (
-                total_loss +
-                (torch.stack(direction_losses) * modality_loss_weights).sum() * self.direction_loss_weight
-            )
-
         # whether to handle velocity consistency
         # for straightening the flow, from consistency flow matching paper https://arxiv.org/abs/2407.02398
 
@@ -2181,7 +2263,7 @@ class Transfusion(Module):
 
             total_loss = (
                 total_loss +
-                (torch.stack(velocity_match_losses) * modality_loss_weights).sum() * self.velocity_consistency_loss_weight
+                (stack(velocity_match_losses) * modality_loss_weights).sum() * self.velocity_consistency_loss_weight
             )
 
         # return total loss if no breakdown needed
@@ -2189,4 +2271,4 @@ class Transfusion(Module):
         if not return_breakdown:
             return total_loss
 
-        return total_loss, LossBreakdown(total_loss, text_loss, flow_losses, velocity_match_losses, direction_losses)
+        return total_loss, LossBreakdown(total_loss, text_loss, flow_losses, velocity_match_losses)
